@@ -1,36 +1,24 @@
-"""
-    sphinx.config
-    ~~~~~~~~~~~~~
-
-    Build configuration file handling.
-
-    :copyright: Copyright 2007-2019 by the Sphinx team, see AUTHORS.
-    :license: BSD, see LICENSE for details.
-"""
+"""Build configuration file handling."""
 
 import re
 import traceback
 import types
-import warnings
 from collections import OrderedDict
-from os import path, getenv
-from typing import Any, NamedTuple, Union
+from os import getenv, path
+from typing import (TYPE_CHECKING, Any, Callable, Dict, Generator, Iterator, List, NamedTuple,
+                    Optional, Set, Tuple, Union)
 
-from sphinx.deprecation import RemovedInSphinx30Warning, RemovedInSphinx40Warning
 from sphinx.errors import ConfigError, ExtensionError
 from sphinx.locale import _, __
 from sphinx.util import logging
 from sphinx.util.i18n import format_date
-from sphinx.util.osutil import cd
-from sphinx.util.pycompat import execfile_
+from sphinx.util.osutil import cd, fs_encoding
+from sphinx.util.tags import Tags
 from sphinx.util.typing import NoneType
 
-if False:
-    # For type annotation
-    from typing import Callable, Dict, Generator, Iterator, List, Set, Tuple  # NOQA
-    from sphinx.application import Sphinx  # NOQA
-    from sphinx.environment import BuildEnvironment  # NOQA
-    from sphinx.util.tags import Tags  # NOQA
+if TYPE_CHECKING:
+    from sphinx.application import Sphinx
+    from sphinx.environment import BuildEnvironment
 
 logger = logging.getLogger(__name__)
 
@@ -38,13 +26,14 @@ CONFIG_FILENAME = 'conf.py'
 UNSERIALIZABLE_TYPES = (type, types.ModuleType, types.FunctionType)
 copyright_year_re = re.compile(r'^((\d{4}-)?)(\d{4})(?=[ ,])')
 
-ConfigValue = NamedTuple('ConfigValue', [('name', str),
-                                         ('value', Any),
-                                         ('rebuild', Union[bool, str])])
+
+class ConfigValue(NamedTuple):
+    name: str
+    value: Any
+    rebuild: Union[bool, str]
 
 
-def is_serializable(obj):
-    # type: (Any) -> bool
+def is_serializable(obj: Any) -> bool:
     """Check if object is serializable or not."""
     if isinstance(obj, UNSERIALIZABLE_TYPES):
         return False
@@ -59,27 +48,21 @@ def is_serializable(obj):
 
 
 class ENUM:
-    """represents the config value should be a one of candidates.
+    """Represents the candidates which a config value should be one of.
 
     Example:
         app.add_config_value('latex_show_urls', 'no', None, ENUM('no', 'footnote', 'inline'))
     """
-    def __init__(self, *candidates):
-        # type: (str) -> None
+    def __init__(self, *candidates: str) -> None:
         self.candidates = candidates
 
-    def match(self, value):
-        # type: (Union[str, List, Tuple]) -> bool
+    def match(self, value: Union[str, List, Tuple]) -> bool:
         if isinstance(value, (list, tuple)):
             return all(item in self.candidates for item in value)
         else:
             return value in self.candidates
 
 
-# RemovedInSphinx40Warning
-string_classes = [str]  # type: List
-
-
 class Config:
     """Configuration file abstraction.
 
@@ -97,25 +80,27 @@ class Config:
     # If you add a value here, don't forget to include it in the
     # quickstart.py file template as well as in the docs!
 
-    config_values = {
+    config_values: Dict[str, Tuple] = {
         # general options
         'project': ('Python', 'env', []),
         'author': ('unknown', 'env', []),
-        'copyright': ('', 'html', []),
+        'project_copyright': ('', 'html', [str]),
+        'copyright': (lambda c: c.project_copyright, 'html', [str]),
         'version': ('', 'env', []),
         'release': ('', 'env', []),
         'today': ('', 'env', []),
         # the real default is locale-dependent
         'today_fmt': (None, 'env', [str]),
 
-        'language': (None, 'env', [str]),
+        'language': ('en', 'env', [str]),
         'locale_dirs': (['locales'], 'env', []),
         'figure_language_filename': ('{root}.{language}{ext}', 'env', [str]),
+        'gettext_allow_fuzzy_translations': (False, 'gettext', []),
 
         'master_doc': ('index', 'env', []),
+        'root_doc': (lambda config: config.master_doc, 'env', []),
         'source_suffix': ({'.rst': 'restructuredtext'}, 'env', Any),
         'source_encoding': ('utf-8-sig', 'env', []),
-        'source_parsers': ({}, 'env', []),
         'exclude_patterns': ([], 'env', []),
         'default_role': (None, 'env', [str]),
         'add_function_parentheses': (True, 'env', []),
@@ -133,12 +118,13 @@ class Config:
         'rst_epilog': (None, 'env', [str]),
         'rst_prolog': (None, 'env', [str]),
         'trim_doctest_flags': (True, 'env', []),
-        'primary_domain': ('py', 'env', [NoneType]),   # type: ignore
+        'primary_domain': ('py', 'env', [NoneType]),
         'needs_sphinx': (None, None, [str]),
         'needs_extensions': ({}, None, []),
         'manpages_url': (None, 'env', []),
         'nitpicky': (False, None, []),
         'nitpick_ignore': ([], None, []),
+        'nitpick_ignore_regex': ([], None, []),
         'numfig': (False, 'env', []),
         'numfig_secnum_depth': (1, 'env', []),
         'numfig_format': ({}, 'env', []),  # will be initialized in init_numfig_format()
@@ -154,69 +140,57 @@ class Config:
         'smartquotes_excludes': ({'languages': ['ja'],
                                   'builders': ['man', 'text']},
                                  'env', []),
-    }  # type: Dict[str, Tuple]
-
-    def __init__(self, *args):
-        # type: (Any) -> None
-        if len(args) == 4:
-            # old style arguments: (dirname, filename, overrides, tags)
-            warnings.warn('The argument of Config() class has been changed. '
-                          'Use Config.read() to read configuration from conf.py.',
-                          RemovedInSphinx30Warning, stacklevel=2)
-            dirname, filename, overrides, tags = args
-            if dirname is None:
-                config = {}  # type: Dict[str, Any]
-            else:
-                config = eval_config_file(path.join(dirname, filename), tags)
-        else:
-            # new style arguments: (config={}, overrides={})
-            if len(args) == 0:
-                config, overrides = {}, {}
-            elif len(args) == 1:
-                config, overrides = args[0], {}
-            else:
-                config, overrides = args[:2]
+    }
 
+    def __init__(self, config: Dict[str, Any] = {}, overrides: Dict[str, Any] = {}) -> None:
         self.overrides = dict(overrides)
         self.values = Config.config_values.copy()
         self._raw_config = config
-        self.setup = config.get('setup', None)  # type: Callable
+        self.setup: Optional[Callable] = config.get('setup', None)
 
         if 'extensions' in self.overrides:
             if isinstance(self.overrides['extensions'], str):
                 config['extensions'] = self.overrides.pop('extensions').split(',')
             else:
                 config['extensions'] = self.overrides.pop('extensions')
-        self.extensions = config.get('extensions', [])  # type: List[str]
+        self.extensions: List[str] = config.get('extensions', [])
 
     @classmethod
-    def read(cls, confdir, overrides=None, tags=None):
-        # type: (str, Dict, Tags) -> Config
+    def read(cls, confdir: str, overrides: Dict = None, tags: Tags = None) -> "Config":
         """Create a Config object from configuration file."""
         filename = path.join(confdir, CONFIG_FILENAME)
+        if not path.isfile(filename):
+            raise ConfigError(__("config directory doesn't contain a conf.py file (%s)") %
+                              confdir)
         namespace = eval_config_file(filename, tags)
+
+        # Note: Old sphinx projects have been configured as "langugae = None" because
+        #       sphinx-quickstart previously generated this by default.
+        #       To keep compatibility, they should be fallback to 'en' for a while
+        #       (This conversion should not be removed before 2025-01-01).
+        if namespace.get("language", ...) is None:
+            logger.warning(__("Invalid configuration value found: 'language = None'. "
+                              "Update your configuration to a valid langauge code. "
+                              "Falling back to 'en' (English)."))
+            namespace["language"] = "en"
+
         return cls(namespace, overrides or {})
 
-    def check_types(self):
-        # type: () -> None
-        warnings.warn('Config.check_types() is deprecated. Use check_confval_types() instead.',
-                      RemovedInSphinx30Warning, stacklevel=2)
-        check_confval_types(None, self)
-
-    def check_unicode(self):
-        # type: () -> None
-        warnings.warn('Config.check_unicode() is deprecated. Use check_unicode() instead.',
-                      RemovedInSphinx30Warning, stacklevel=2)
-        check_unicode(self)
-
-    def convert_overrides(self, name, value):
-        # type: (str, Any) -> Any
+    def convert_overrides(self, name: str, value: Any) -> Any:
         if not isinstance(value, str):
             return value
         else:
             defvalue = self.values[name][0]
             if self.values[name][2] == Any:
                 return value
+            elif self.values[name][2] == {bool, str}:
+                if value == '0':
+                    # given falsy string from command line option
+                    return False
+                elif value == '1':
+                    return True
+                else:
+                    return value
             elif type(defvalue) is bool or self.values[name][2] == [bool]:
                 if value == '0':
                     # given falsy string from command line option
@@ -232,10 +206,10 @@ class Config:
             elif isinstance(defvalue, int):
                 try:
                     return int(value)
-                except ValueError:
+                except ValueError as exc:
                     raise ValueError(__('invalid number %r for config value %r, ignoring') %
-                                     (value, name))
-            elif hasattr(defvalue, '__call__'):
+                                     (value, name)) from exc
+            elif callable(defvalue):
                 return value
             elif defvalue is not None and not isinstance(defvalue, str):
                 raise ValueError(__('cannot override config setting %r with unsupported '
@@ -243,10 +217,10 @@ class Config:
             else:
                 return value
 
-    def pre_init_values(self):
-        # type: () -> None
+    def pre_init_values(self) -> None:
         """
-        Initialize some limited config variables before initialize i18n and loading extensions
+        Initialize some limited config variables before initializing i18n and loading
+        extensions.
         """
         variables = ['needs_sphinx', 'suppress_warnings', 'language', 'locale_dirs']
         for name in variables:
@@ -258,8 +232,7 @@ class Config:
             except ValueError as exc:
                 logger.warning("%s", exc)
 
-    def init_values(self):
-        # type: () -> None
+    def init_values(self) -> None:
         config = self._raw_config
         for valname, value in self.overrides.items():
             try:
@@ -281,53 +254,55 @@ class Config:
             if name in self.values:
                 self.__dict__[name] = config[name]
 
-    def __getattr__(self, name):
-        # type: (str) -> Any
+    def post_init_values(self) -> None:
+        """
+        Initialize additional config variables that are added after init_values() called.
+        """
+        config = self._raw_config
+        for name in config:
+            if name not in self.__dict__ and name in self.values:
+                self.__dict__[name] = config[name]
+
+        check_confval_types(None, self)
+
+    def __getattr__(self, name: str) -> Any:
         if name.startswith('_'):
             raise AttributeError(name)
         if name not in self.values:
             raise AttributeError(__('No such config value: %s') % name)
         default = self.values[name][0]
-        if hasattr(default, '__call__'):
+        if callable(default):
             return default(self)
         return default
 
-    def __getitem__(self, name):
-        # type: (str) -> str
+    def __getitem__(self, name: str) -> Any:
         return getattr(self, name)
 
-    def __setitem__(self, name, value):
-        # type: (str, Any) -> None
+    def __setitem__(self, name: str, value: Any) -> None:
         setattr(self, name, value)
 
-    def __delitem__(self, name):
-        # type: (str) -> None
+    def __delitem__(self, name: str) -> None:
         delattr(self, name)
 
-    def __contains__(self, name):
-        # type: (str) -> bool
+    def __contains__(self, name: str) -> bool:
         return name in self.values
 
-    def __iter__(self):
-        # type: () -> Generator[ConfigValue, None, None]
+    def __iter__(self) -> Generator[ConfigValue, None, None]:
         for name, value in self.values.items():
             yield ConfigValue(name, getattr(self, name), value[1])
 
-    def add(self, name, default, rebuild, types):
-        # type: (str, Any, Union[bool, str], Any) -> None
+    def add(self, name: str, default: Any, rebuild: Union[bool, str], types: Any) -> None:
         if name in self.values:
             raise ExtensionError(__('Config value %r already present') % name)
         else:
             self.values[name] = (default, rebuild, types)
 
-    def filter(self, rebuild):
-        # type: (Union[str, List[str]]) -> Iterator[ConfigValue]
+    def filter(self, rebuild: Union[str, List[str]]) -> Iterator[ConfigValue]:
         if isinstance(rebuild, str):
             rebuild = [rebuild]
         return (value for value in self if value.rebuild in rebuild)
 
-    def __getstate__(self):
-        # type: () -> Dict
+    def __getstate__(self) -> Dict:
         """Obtains serializable data for pickling."""
         # remove potentially pickling-problematic values from config
         __dict__ = {}
@@ -350,39 +325,41 @@ class Config:
 
         return __dict__
 
-    def __setstate__(self, state):
-        # type: (Dict) -> None
+    def __setstate__(self, state: Dict) -> None:
         self.__dict__.update(state)
 
 
-def eval_config_file(filename, tags):
-    # type: (str, Tags) -> Dict[str, Any]
+def eval_config_file(filename: str, tags: Optional[Tags]) -> Dict[str, Any]:
     """Evaluate a config file."""
-    namespace = {}  # type: Dict[str, Any]
+    namespace: Dict[str, Any] = {}
     namespace['__file__'] = filename
     namespace['tags'] = tags
 
     with cd(path.dirname(filename)):
         # during executing config file, current dir is changed to ``confdir``.
         try:
-            execfile_(filename, namespace)
+            with open(filename, 'rb') as f:
+                code = compile(f.read(), filename.encode(fs_encoding), 'exec')
+                exec(code, namespace)
         except SyntaxError as err:
             msg = __("There is a syntax error in your configuration file: %s\n")
-            raise ConfigError(msg % err)
-        except SystemExit:
+            raise ConfigError(msg % err) from err
+        except SystemExit as exc:
             msg = __("The configuration file (or one of the modules it imports) "
                      "called sys.exit()")
-            raise ConfigError(msg)
-        except Exception:
+            raise ConfigError(msg) from exc
+        except ConfigError:
+            # pass through ConfigError from conf.py as is.  It will be shown in console.
+            raise
+        except Exception as exc:
             msg = __("There is a programmable error in your configuration file:\n\n%s")
-            raise ConfigError(msg % traceback.format_exc())
+            raise ConfigError(msg % traceback.format_exc()) from exc
 
     return namespace
 
 
-def convert_source_suffix(app, config):
-    # type: (Sphinx, Config) -> None
-    """This converts old styled source_suffix to new styled one.
+def convert_source_suffix(app: "Sphinx", config: Config) -> None:
+    """Convert old styled source_suffix to new styled one.
 
     * old style: str or list
     * new style: a dict which maps from fileext to filetype
@@ -406,8 +383,19 @@ def convert_source_suffix(app, config):
                           "But `%r' is given." % source_suffix))
 
 
-def init_numfig_format(app, config):
-    # type: (Sphinx, Config) -> None
+def convert_highlight_options(app: "Sphinx", config: Config) -> None:
+    """Convert old styled highlight_options to new styled one.
+
+    * old style: options
+    * new style: a dict which maps from language name to options
+    """
+    options = config.highlight_options
+    if options and not all(isinstance(v, dict) for v in options.values()):
+        # old styled option detected because all values are not dictionary.
+        config.highlight_options = {config.highlight_language: options}  # type: ignore
+
+
+def init_numfig_format(app: "Sphinx", config: Config) -> None:
     """Initialize :confval:`numfig_format`."""
     numfig_format = {'section': _('Section %s'),
                      'figure': _('Fig. %s'),
@@ -419,9 +407,8 @@ def init_numfig_format(app, config):
     config.numfig_format = numfig_format  # type: ignore
 
 
-def correct_copyright_year(app, config):
-    # type: (Sphinx, Config) -> None
-    """correct values of copyright year that are not coherent with
+def correct_copyright_year(app: "Sphinx", config: Config) -> None:
+    """Correct values of copyright year that are not coherent with
     the SOURCE_DATE_EPOCH environment variable (if set)
 
     See https://reproducible-builds.org/specs/source-date-epoch/
@@ -429,22 +416,21 @@ def correct_copyright_year(app, config):
     if getenv('SOURCE_DATE_EPOCH') is not None:
         for k in ('copyright', 'epub_copyright'):
             if k in config:
-                replace = r'\g<1>%s' % format_date('%Y')
+                replace = r'\g<1>%s' % format_date('%Y', language='en')
                 config[k] = copyright_year_re.sub(replace, config[k])
 
 
-def check_confval_types(app, config):
-    # type: (Sphinx, Config) -> None
-    """check all values for deviation from the default value's type, since
+def check_confval_types(app: "Sphinx", config: Config) -> None:
+    """Check all values for deviation from the default value's type, since
     that can result in TypeErrors all over the place NB.
     """
     for confval in config:
         default, rebuild, annotations = config.values[confval.name]
 
-        if hasattr(default, '__call__'):
+        if callable(default):
             default = default(config)  # evaluate default value
         if default is None and not annotations:
-            continue  # neither inferrable nor expliclitly annotated types
+            continue  # neither inferable nor expliclitly annotated types
 
         if annotations is Any:
             # any type of value is accepted
@@ -455,7 +441,7 @@ def check_confval_types(app, config):
                          "but `{current}` is given.")
                 logger.warning(msg.format(name=confval.name,
                                           current=confval.value,
-                                          candidates=annotations.candidates))
+                                          candidates=annotations.candidates), once=True)
         else:
             if type(confval.value) is type(default):
                 continue
@@ -480,63 +466,45 @@ def check_confval_types(app, config):
                     permitted = " or ".join(wrapped_annotations)
                 logger.warning(msg.format(name=confval.name,
                                           current=type(confval.value),
-                                          permitted=permitted))
+                                          permitted=permitted), once=True)
             else:
                 msg = __("The config value `{name}' has type `{current.__name__}', "
                          "defaults to `{default.__name__}'.")
                 logger.warning(msg.format(name=confval.name,
                                           current=type(confval.value),
-                                          default=type(default)))
+                                          default=type(default)), once=True)
 
 
-def check_unicode(config):
-    # type: (Config) -> None
-    """check all string values for non-ASCII characters in bytestrings,
-    since that can result in UnicodeErrors all over the place
-    """
-    warnings.warn('sphinx.config.check_unicode() is deprecated.',
-                  RemovedInSphinx40Warning)
-
-    nonascii_re = re.compile(br'[\x80-\xff]')
-
-    for name, value in config._raw_config.items():
-        if isinstance(value, bytes) and nonascii_re.search(value):
-            logger.warning(__('the config value %r is set to a string with non-ASCII '
-                              'characters; this can lead to Unicode errors occurring. '
-                              'Please use Unicode strings, e.g. %r.'), name, 'Content')
-
-
-def check_primary_domain(app, config):
-    # type: (Sphinx, Config) -> None
+def check_primary_domain(app: "Sphinx", config: Config) -> None:
     primary_domain = config.primary_domain
     if primary_domain and not app.registry.has_domain(primary_domain):
         logger.warning(__('primary_domain %r not found, ignored.'), primary_domain)
         config.primary_domain = None  # type: ignore
 
 
-def check_master_doc(app, env, added, changed, removed):
-    # type: (Sphinx, BuildEnvironment, Set[str], Set[str], Set[str]) -> Set[str]
-    """Adjust master_doc to 'contents' to support an old project which does not have
-    no master_doc setting.
+def check_root_doc(app: "Sphinx", env: "BuildEnvironment", added: Set[str],
+                   changed: Set[str], removed: Set[str]) -> Set[str]:
+    """Adjust root_doc to 'contents' to support an old project which does not have
+    any root_doc setting.
     """
-    if (app.config.master_doc == 'index' and
+    if (app.config.root_doc == 'index' and
             'index' not in app.project.docnames and
             'contents' in app.project.docnames):
-        logger.warning(__('Since v2.0, Sphinx uses "index" as master_doc by default. '
-                          'Please add "master_doc = \'contents\'" to your conf.py.'))
-        app.config.master_doc = "contents"  # type: ignore
+        logger.warning(__('Since v2.0, Sphinx uses "index" as root_doc by default. '
+                          'Please add "root_doc = \'contents\'" to your conf.py.'))
+        app.config.root_doc = "contents"  # type: ignore
 
     return changed
 
 
-def setup(app):
-    # type: (Sphinx) -> Dict[str, Any]
-    app.connect('config-inited', convert_source_suffix)
-    app.connect('config-inited', init_numfig_format)
-    app.connect('config-inited', correct_copyright_year)
-    app.connect('config-inited', check_confval_types)
-    app.connect('config-inited', check_primary_domain)
-    app.connect('env-get-outdated', check_master_doc)
+def setup(app: "Sphinx") -> Dict[str, Any]:
+    app.connect('config-inited', convert_source_suffix, priority=800)
+    app.connect('config-inited', convert_highlight_options, priority=800)
+    app.connect('config-inited', init_numfig_format, priority=800)
+    app.connect('config-inited', correct_copyright_year, priority=800)
+    app.connect('config-inited', check_confval_types, priority=800)
+    app.connect('config-inited', check_primary_domain, priority=800)
+    app.connect('env-get-outdated', check_root_doc)
 
     return {
         'version': 'builtin',
