@@ -1,50 +1,44 @@
-"""
-    sphinx.registry
-    ~~~~~~~~~~~~~~~
-
-    Sphinx component registry.
-
-    :copyright: Copyright 2007-2016 by the Sphinx team, see AUTHORS.
-    :license: BSD, see LICENSE for details.
-"""
+"""Sphinx component registry."""
 
 import traceback
 import warnings
 from importlib import import_module
-from inspect import isclass
 from types import MethodType
+from typing import (TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Tuple, Type,
+                    Union)
 
+from docutils import nodes
+from docutils.core import Publisher
+from docutils.io import Input
+from docutils.nodes import Element, Node, TextElement
+from docutils.parsers import Parser
 from docutils.parsers.rst import Directive
-from pkg_resources import iter_entry_points
+from docutils.transforms import Transform
 
-from sphinx.deprecation import RemovedInSphinx30Warning
-from sphinx.domains import ObjType
+try:  # Python < 3.10 (backport)
+    from importlib_metadata import entry_points
+except ImportError:
+    from importlib.metadata import entry_points
+
+from sphinx.builders import Builder
+from sphinx.config import Config
+from sphinx.deprecation import RemovedInSphinx60Warning
+from sphinx.domains import Domain, Index, ObjType
 from sphinx.domains.std import GenericObject, Target
+from sphinx.environment import BuildEnvironment
 from sphinx.errors import ExtensionError, SphinxError, VersionRequirementError
 from sphinx.extension import Extension
+from sphinx.io import create_publisher
 from sphinx.locale import __
 from sphinx.parsers import Parser as SphinxParser
 from sphinx.roles import XRefRole
 from sphinx.util import logging
-from sphinx.util.docutils import directive_helper
 from sphinx.util.logging import prefixed_warnings
+from sphinx.util.typing import RoleFunction, TitleGetter
 
-if False:
-    # For type annotation
-    from typing import Any, Callable, Dict, Iterator, List, Tuple, Union  # NOQA
-    from typing import Type  # for python3.5.1
-    from docutils import nodes  # NOQA
-    from docutils.io import Input  # NOQA
-    from docutils.parsers import Parser  # NOQA
-    from docutils.transforms import Transform  # NOQA
-    from sphinx.application import Sphinx  # NOQA
-    from sphinx.builders import Builder  # NOQA
-    from sphinx.config import Config  # NOQA
-    from sphinx.domains import Domain, Index  # NOQA
-    from sphinx.environment import BuildEnvironment  # NOQA
-    from sphinx.ext.autodoc import Documenter  # NOQA
-    from sphinx.io import SphinxFileInput  # NOQA
-    from sphinx.util.typing import RoleFunction, TitleGetter  # NOQA
+if TYPE_CHECKING:
+    from sphinx.application import Sphinx
+    from sphinx.ext.autodoc import Documenter
 
 logger = logging.getLogger(__name__)
 
@@ -56,78 +50,87 @@ EXTENSION_BLACKLIST = {
 
 
 class SphinxComponentRegistry:
-    def __init__(self):
-        # type: () -> None
+    def __init__(self) -> None:
         #: special attrgetter for autodoc; class object -> attrgetter
-        self.autodoc_attrgettrs = {}    # type: Dict[Type, Callable[[Any, str, Any], Any]]
+        self.autodoc_attrgettrs: Dict[Type, Callable[[Any, str, Any], Any]] = {}
 
         #: builders; a dict of builder name -> bulider class
-        self.builders = {}              # type: Dict[str, Type[Builder]]
+        self.builders: Dict[str, Type[Builder]] = {}
 
         #: autodoc documenters; a dict of documenter name -> documenter class
-        self.documenters = {}           # type: Dict[str, Type[Documenter]]
+        self.documenters: Dict[str, Type[Documenter]] = {}
 
         #: css_files; a list of tuple of filename and attributes
-        self.css_files = []             # type: List[Tuple[str, Dict[str, str]]]
+        self.css_files: List[Tuple[str, Dict[str, Any]]] = []
 
         #: domains; a dict of domain name -> domain class
-        self.domains = {}               # type: Dict[str, Type[Domain]]
+        self.domains: Dict[str, Type[Domain]] = {}
 
         #: additional directives for domains
         #: a dict of domain name -> dict of directive name -> directive
-        self.domain_directives = {}     # type: Dict[str, Dict[str, Any]]
+        self.domain_directives: Dict[str, Dict[str, Any]] = {}
 
         #: additional indices for domains
         #: a dict of domain name -> list of index class
-        self.domain_indices = {}        # type: Dict[str, List[Type[Index]]]
+        self.domain_indices: Dict[str, List[Type[Index]]] = {}
 
         #: additional object types for domains
         #: a dict of domain name -> dict of objtype name -> objtype
-        self.domain_object_types = {}   # type: Dict[str, Dict[str, ObjType]]
+        self.domain_object_types: Dict[str, Dict[str, ObjType]] = {}
 
         #: additional roles for domains
         #: a dict of domain name -> dict of role name -> role impl.
-        self.domain_roles = {}          # type: Dict[str, Dict[str, Union[RoleFunction, XRefRole]]]  # NOQA
+        self.domain_roles: Dict[str, Dict[str, Union[RoleFunction, XRefRole]]] = {}
 
         #: additional enumerable nodes
         #: a dict of node class -> tuple of figtype and title_getter function
-        self.enumerable_nodes = {}      # type: Dict[Type[nodes.Node], Tuple[str, TitleGetter]]
+        self.enumerable_nodes: Dict[Type[Node], Tuple[str, TitleGetter]] = {}
 
         #: HTML inline and block math renderers
         #: a dict of name -> tuple of visit function and depart function
-        self.html_inline_math_renderers = {}    # type: Dict[str, Tuple[Callable, Callable]]
-        self.html_block_math_renderers = {}     # type: Dict[str, Tuple[Callable, Callable]]
+        self.html_inline_math_renderers: Dict[str, Tuple[Callable, Callable]] = {}
+        self.html_block_math_renderers: Dict[str, Tuple[Callable, Callable]] = {}
+
+        #: HTML assets
+        self.html_assets_policy: str = 'per_page'
+
+        #: HTML themes
+        self.html_themes: Dict[str, str] = {}
 
         #: js_files; list of JS paths or URLs
-        self.js_files = []              # type: List[Tuple[str, Dict[str, str]]]
+        self.js_files: List[Tuple[str, Dict[str, Any]]] = []
 
         #: LaTeX packages; list of package names and its options
-        self.latex_packages = []        # type: List[Tuple[str, str]]
+        self.latex_packages: List[Tuple[str, str]] = []
+
+        self.latex_packages_after_hyperref: List[Tuple[str, str]] = []
 
         #: post transforms; list of transforms
-        self.post_transforms = []       # type: List[Type[Transform]]
+        self.post_transforms: List[Type[Transform]] = []
 
         #: source paresrs; file type -> parser class
-        self.source_parsers = {}        # type: Dict[str, Type[Parser]]
+        self.source_parsers: Dict[str, Type[Parser]] = {}
 
         #: source inputs; file type -> input class
-        self.source_inputs = {}         # type: Dict[str, Type[Input]]
+        self.source_inputs: Dict[str, Type[Input]] = {}
 
         #: source suffix: suffix -> file type
-        self.source_suffix = {}         # type: Dict[str, str]
+        self.source_suffix: Dict[str, str] = {}
 
         #: custom translators; builder name -> translator class
-        self.translators = {}           # type: Dict[str, Type[nodes.NodeVisitor]]
+        self.translators: Dict[str, Type[nodes.NodeVisitor]] = {}
 
         #: custom handlers for translators
         #: a dict of builder name -> dict of node name -> visitor and departure functions
-        self.translation_handlers = {}  # type: Dict[str, Dict[str, Tuple[Callable, Callable]]]
+        self.translation_handlers: Dict[str, Dict[str, Tuple[Callable, Callable]]] = {}
 
         #: additional transforms; list of transforms
-        self.transforms = []            # type: List[Type[Transform]]
+        self.transforms: List[Type[Transform]] = []
 
-    def add_builder(self, builder, override=False):
-        # type: (Type[Builder], bool) -> None
+        # private cache of Docutils Publishers (file type -> publisher object)
+        self.publishers: Dict[str, Publisher] = {}
+
+    def add_builder(self, builder: Type[Builder], override: bool = False) -> None:
         logger.debug('[app] adding builder: %r', builder)
         if not hasattr(builder, 'name'):
             raise ExtensionError(__('Builder class %s has no "name" attribute') % builder)
@@ -136,41 +139,36 @@ class SphinxComponentRegistry:
                                  (builder.name, self.builders[builder.name].__module__))
         self.builders[builder.name] = builder
 
-    def preload_builder(self, app, name):
-        # type: (Sphinx, str) -> None
+    def preload_builder(self, app: "Sphinx", name: str) -> None:
         if name is None:
             return
 
         if name not in self.builders:
-            entry_points = iter_entry_points('sphinx.builders', name)
+            builder_entry_points = entry_points(group='sphinx.builders')
             try:
-                entry_point = next(entry_points)
-            except StopIteration:
+                entry_point = builder_entry_points[name]
+            except KeyError as exc:
                 raise SphinxError(__('Builder name %s not registered or available'
-                                     ' through entry point') % name)
+                                     ' through entry point') % name) from exc
 
-            self.load_extension(app, entry_point.module_name)
+            self.load_extension(app, entry_point.module)
 
-    def create_builder(self, app, name):
-        # type: (Sphinx, str) -> Builder
+    def create_builder(self, app: "Sphinx", name: str) -> Builder:
         if name not in self.builders:
             raise SphinxError(__('Builder name %s not registered') % name)
 
         return self.builders[name](app)
 
-    def add_domain(self, domain, override=False):
-        # type: (Type[Domain], bool) -> None
+    def add_domain(self, domain: Type[Domain], override: bool = False) -> None:
         logger.debug('[app] adding domain: %r', domain)
         if domain.name in self.domains and not override:
             raise ExtensionError(__('domain %s already registered') % domain.name)
         self.domains[domain.name] = domain
 
-    def has_domain(self, domain):
-        # type: (str) -> bool
+    def has_domain(self, domain: str) -> bool:
         return domain in self.domains
 
-    def create_domains(self, env):
-        # type: (BuildEnvironment) -> Iterator[Domain]
+    def create_domains(self, env: BuildEnvironment) -> Iterator[Domain]:
         for DomainClass in self.domains.values():
             domain = DomainClass(env)
 
@@ -183,18 +181,9 @@ class SphinxComponentRegistry:
 
             yield domain
 
-    def override_domain(self, domain):
-        # type: (Type[Domain]) -> None
-        warnings.warn('registry.override_domain() is deprecated. '
-                      'Use app.add_domain(domain, override=True) instead.',
-                      RemovedInSphinx30Warning, stacklevel=2)
-        self.add_domain(domain, override=True)
-
-    def add_directive_to_domain(self, domain, name, obj, has_content=None, argument_spec=None,
-                                override=False, **option_spec):
-        # type: (str, str, Any, bool, Any, bool, Any) -> None
-        logger.debug('[app] adding directive to domain: %r',
-                     (domain, name, obj, has_content, argument_spec, option_spec))
+    def add_directive_to_domain(self, domain: str, name: str,
+                                cls: Type[Directive], override: bool = False) -> None:
+        logger.debug('[app] adding directive to domain: %r', (domain, name, cls))
         if domain not in self.domains:
             raise ExtensionError(__('domain %s not yet registered') % domain)
 
@@ -202,13 +191,11 @@ class SphinxComponentRegistry:
         if name in directives and not override:
             raise ExtensionError(__('The %r directive is already registered to domain %s') %
                                  (name, domain))
-        if not isclass(obj) or not issubclass(obj, Directive):
-            directives[name] = directive_helper(obj, has_content, argument_spec, **option_spec)
-        else:
-            directives[name] = obj
+        directives[name] = cls
 
-    def add_role_to_domain(self, domain, name, role, override=False):
-        # type: (str, str, Union[RoleFunction, XRefRole], bool) -> None
+    def add_role_to_domain(self, domain: str, name: str,
+                           role: Union[RoleFunction, XRefRole], override: bool = False
+                           ) -> None:
         logger.debug('[app] adding role to domain: %r', (domain, name, role))
         if domain not in self.domains:
             raise ExtensionError(__('domain %s not yet registered') % domain)
@@ -218,8 +205,8 @@ class SphinxComponentRegistry:
                                  (name, domain))
         roles[name] = role
 
-    def add_index_to_domain(self, domain, index, override=False):
-        # type: (str, Type[Index], bool) -> None
+    def add_index_to_domain(self, domain: str, index: Type[Index],
+                            override: bool = False) -> None:
         logger.debug('[app] adding index to domain: %r', (domain, index))
         if domain not in self.domains:
             raise ExtensionError(__('domain %s not yet registered') % domain)
@@ -229,10 +216,10 @@ class SphinxComponentRegistry:
                                  (index.name, domain))
         indices.append(index)
 
-    def add_object_type(self, directivename, rolename, indextemplate='',
-                        parse_node=None, ref_nodeclass=None, objname='',
-                        doc_field_types=[], override=False):
-        # type: (str, str, str, Callable, Type[nodes.TextElement], str, List, bool) -> None
+    def add_object_type(self, directivename: str, rolename: str, indextemplate: str = '',
+                        parse_node: Callable = None, ref_nodeclass: Type[TextElement] = None,
+                        objname: str = '', doc_field_types: List = [], override: bool = False
+                        ) -> None:
         logger.debug('[app] adding object type: %r',
                      (directivename, rolename, indextemplate, parse_node,
                       ref_nodeclass, objname, doc_field_types))
@@ -253,9 +240,9 @@ class SphinxComponentRegistry:
                                  directivename)
         object_types[directivename] = ObjType(objname or directivename, rolename)
 
-    def add_crossref_type(self, directivename, rolename, indextemplate='',
-                          ref_nodeclass=None, objname='', override=False):
-        # type: (str, str, str, Type[nodes.TextElement], str, bool) -> None
+    def add_crossref_type(self, directivename: str, rolename: str, indextemplate: str = '',
+                          ref_nodeclass: Type[TextElement] = None, objname: str = '',
+                          override: bool = False) -> None:
         logger.debug('[app] adding crossref type: %r',
                      (directivename, rolename, indextemplate, ref_nodeclass, objname))
 
@@ -273,82 +260,44 @@ class SphinxComponentRegistry:
                                  directivename)
         object_types[directivename] = ObjType(objname or directivename, rolename)
 
-    def add_source_suffix(self, suffix, filetype, override=False):
-        # type: (str, str, bool) -> None
+    def add_source_suffix(self, suffix: str, filetype: str, override: bool = False) -> None:
         logger.debug('[app] adding source_suffix: %r, %r', suffix, filetype)
         if suffix in self.source_suffix and not override:
             raise ExtensionError(__('source_suffix %r is already registered') % suffix)
         else:
             self.source_suffix[suffix] = filetype
 
-    def add_source_parser(self, *args, **kwargs):
-        # type: (Any, bool) -> None
-        logger.debug('[app] adding search source_parser: %r', args)
-        if len(args) == 1:
-            # new sytle arguments: (source_parser)
-            suffix = None       # type: str
-            parser = args[0]    # type: Type[Parser]
-        else:
-            # old style arguments: (suffix, source_parser)
-            warnings.warn('app.add_source_parser() does not support suffix argument. '
-                          'Use app.add_source_suffix() instead.',
-                          RemovedInSphinx30Warning, stacklevel=3)
-            suffix = args[0]
-            parser = args[1]
-
-        if suffix:
-            self.add_source_suffix(suffix, suffix, override=True)
-
-        if len(parser.supported) == 0:
-            warnings.warn('Old source_parser has been detected. Please fill Parser.supported '
-                          'attribute: %s' % parser.__name__,
-                          RemovedInSphinx30Warning, stacklevel=3)
+    def add_source_parser(self, parser: Type[Parser], override: bool = False) -> None:
+        logger.debug('[app] adding search source_parser: %r', parser)
 
         # create a map from filetype to parser
         for filetype in parser.supported:
-            if filetype in self.source_parsers and not kwargs.get('override'):
+            if filetype in self.source_parsers and not override:
                 raise ExtensionError(__('source_parser for %r is already registered') %
                                      filetype)
             else:
                 self.source_parsers[filetype] = parser
 
-        # also maps suffix to parser
-        #
-        # This rescues old styled parsers which does not have ``supported`` filetypes.
-        if suffix:
-            self.source_parsers[suffix] = parser
-
-    def get_source_parser(self, filetype):
-        # type: (str) -> Type[Parser]
+    def get_source_parser(self, filetype: str) -> Type[Parser]:
         try:
             return self.source_parsers[filetype]
-        except KeyError:
-            raise SphinxError(__('Source parser for %s not registered') % filetype)
+        except KeyError as exc:
+            raise SphinxError(__('Source parser for %s not registered') % filetype) from exc
 
-    def get_source_parsers(self):
-        # type: () -> Dict[str, Type[Parser]]
+    def get_source_parsers(self) -> Dict[str, Type[Parser]]:
         return self.source_parsers
 
-    def create_source_parser(self, app, filename):
-        # type: (Sphinx, str) -> Parser
+    def create_source_parser(self, app: "Sphinx", filename: str) -> Parser:
         parser_class = self.get_source_parser(filename)
         parser = parser_class()
         if isinstance(parser, SphinxParser):
             parser.set_application(app)
         return parser
 
-    def add_source_input(self, input_class, override=False):
-        # type: (Type[SphinxFileInput], bool) -> None
-        warnings.warn('registry.source_input() is deprecated.',
-                      RemovedInSphinx30Warning, stacklevel=2)
-        for filetype in input_class.supported:
-            if filetype in self.source_inputs and not override:
-                raise ExtensionError(__('source_input for %r is already registered') %
-                                     filetype)
-            self.source_inputs[filetype] = input_class
+    def get_source_input(self, filetype: str) -> Optional[Type[Input]]:
+        warnings.warn('SphinxComponentRegistry.get_source_input() is deprecated.',
+                      RemovedInSphinx60Warning)
 
-    def get_source_input(self, filetype):
-        # type: (str) -> Type[Input]
         try:
             return self.source_inputs[filetype]
         except KeyError:
@@ -358,32 +307,32 @@ class SphinxComponentRegistry:
             except KeyError:
                 return None
 
-    def add_translator(self, name, translator, override=False):
-        # type: (str, Type[nodes.NodeVisitor], bool) -> None
-        logger.debug('[app] Change of translator for the %s builder.' % name)
+    def add_translator(self, name: str, translator: Type[nodes.NodeVisitor],
+                       override: bool = False) -> None:
+        logger.debug('[app] Change of translator for the %s builder.', name)
         if name in self.translators and not override:
             raise ExtensionError(__('Translator for %r already exists') % name)
         self.translators[name] = translator
 
-    def add_translation_handlers(self, node, **kwargs):
-        # type: (Type[nodes.Element], Any) -> None
+    def add_translation_handlers(self, node: Type[Element],
+                                 **kwargs: Tuple[Callable, Callable]) -> None:
         logger.debug('[app] adding translation_handlers: %r, %r', node, kwargs)
         for builder_name, handlers in kwargs.items():
             translation_handlers = self.translation_handlers.setdefault(builder_name, {})
             try:
                 visit, depart = handlers  # unpack once for assertion
                 translation_handlers[node.__name__] = (visit, depart)
-            except ValueError:
-                raise ExtensionError(__('kwargs for add_node() must be a (visit, depart) '
-                                        'function tuple: %r=%r') % builder_name, handlers)
+            except ValueError as exc:
+                raise ExtensionError(
+                    __('kwargs for add_node() must be a (visit, depart) '
+                       'function tuple: %r=%r') % (builder_name, handlers)
+                ) from exc
 
-    def get_translator_class(self, builder):
-        # type: (Builder) -> Type[nodes.NodeVisitor]
+    def get_translator_class(self, builder: Builder) -> Type[nodes.NodeVisitor]:
         return self.translators.get(builder.name,
                                     builder.default_translator_class)
 
-    def create_translator(self, builder, *args):
-        # type: (Builder, Any) -> nodes.NodeVisitor
+    def create_translator(self, builder: Builder, *args: Any) -> nodes.NodeVisitor:
         translator_class = self.get_translator_class(builder)
         assert translator_class, "translator not found for %s" % builder.name
         translator = translator_class(*args)
@@ -401,66 +350,72 @@ class SphinxComponentRegistry:
 
         return translator
 
-    def add_transform(self, transform):
-        # type: (Type[Transform]) -> None
+    def add_transform(self, transform: Type[Transform]) -> None:
         logger.debug('[app] adding transform: %r', transform)
         self.transforms.append(transform)
 
-    def get_transforms(self):
-        # type: () -> List[Type[Transform]]
+    def get_transforms(self) -> List[Type[Transform]]:
         return self.transforms
 
-    def add_post_transform(self, transform):
-        # type: (Type[Transform]) -> None
+    def add_post_transform(self, transform: Type[Transform]) -> None:
         logger.debug('[app] adding post transform: %r', transform)
         self.post_transforms.append(transform)
 
-    def get_post_transforms(self):
-        # type: () -> List[Type[Transform]]
+    def get_post_transforms(self) -> List[Type[Transform]]:
         return self.post_transforms
 
-    def add_documenter(self, objtype, documenter):
-        # type: (str, Type[Documenter]) -> None
+    def add_documenter(self, objtype: str, documenter: Type["Documenter"]) -> None:
         self.documenters[objtype] = documenter
 
-    def add_autodoc_attrgetter(self, typ, attrgetter):
-        # type: (Type, Callable[[Any, str, Any], Any]) -> None
+    def add_autodoc_attrgetter(self, typ: Type,
+                               attrgetter: Callable[[Any, str, Any], Any]) -> None:
         self.autodoc_attrgettrs[typ] = attrgetter
 
-    def add_css_files(self, filename, **attributes):
+    def add_css_files(self, filename: str, **attributes: Any) -> None:
         self.css_files.append((filename, attributes))
 
-    def add_js_file(self, filename, **attributes):
-        # type: (str, **str) -> None
+    def add_js_file(self, filename: str, **attributes: Any) -> None:
         logger.debug('[app] adding js_file: %r, %r', filename, attributes)
         self.js_files.append((filename, attributes))
 
-    def add_latex_package(self, name, options):
-        # type: (str, str) -> None
-        logger.debug('[app] adding latex package: %r', name)
-        self.latex_packages.append((name, options))
+    def has_latex_package(self, name: str) -> bool:
+        packages = self.latex_packages + self.latex_packages_after_hyperref
+        return bool([x for x in packages if x[0] == name])
 
-    def add_enumerable_node(self, node, figtype, title_getter=None, override=False):
-        # type: (Type[nodes.Node], str, TitleGetter, bool) -> None
+    def add_latex_package(self, name: str, options: str, after_hyperref: bool = False) -> None:
+        if self.has_latex_package(name):
+            logger.warn("latex package '%s' already included" % name)
+
+        logger.debug('[app] adding latex package: %r', name)
+        if after_hyperref:
+            self.latex_packages_after_hyperref.append((name, options))
+        else:
+            self.latex_packages.append((name, options))
+
+    def add_enumerable_node(self, node: Type[Node], figtype: str,
+                            title_getter: TitleGetter = None, override: bool = False) -> None:
         logger.debug('[app] adding enumerable node: (%r, %r, %r)', node, figtype, title_getter)
         if node in self.enumerable_nodes and not override:
             raise ExtensionError(__('enumerable_node %r already registered') % node)
         self.enumerable_nodes[node] = (figtype, title_getter)
 
-    def add_html_math_renderer(self, name, inline_renderers, block_renderers):
-        # type: (str, Tuple[Callable, Callable], Tuple[Callable, Callable]) -> None
+    def add_html_math_renderer(self, name: str,
+                               inline_renderers: Tuple[Callable, Callable],
+                               block_renderers: Tuple[Callable, Callable]) -> None:
         logger.debug('[app] adding html_math_renderer: %s, %r, %r',
                      name, inline_renderers, block_renderers)
         if name in self.html_inline_math_renderers:
-            raise ExtensionError(__('math renderer %s is already registred') % name)
+            raise ExtensionError(__('math renderer %s is already registered') % name)
 
         self.html_inline_math_renderers[name] = inline_renderers
         self.html_block_math_renderers[name] = block_renderers
 
-    def load_extension(self, app, extname):
-        # type: (Sphinx, str) -> None
+    def add_html_theme(self, name: str, theme_path: str) -> None:
+        self.html_themes[name] = theme_path
+
+    def load_extension(self, app: "Sphinx", extname: str) -> None:
         """Load a Sphinx extension."""
-        if extname in app.extensions:  # alread loaded
+        if extname in app.extensions:  # already loaded
             return
         if extname in EXTENSION_BLACKLIST:
             logger.warning(__('the extension %r was already merged with Sphinx since '
@@ -475,13 +430,14 @@ class SphinxComponentRegistry:
                 mod = import_module(extname)
             except ImportError as err:
                 logger.verbose(__('Original exception:\n') + traceback.format_exc())
-                raise ExtensionError(__('Could not import extension %s') % extname, err)
+                raise ExtensionError(__('Could not import extension %s') % extname,
+                                     err) from err
 
             setup = getattr(mod, 'setup', None)
             if setup is None:
                 logger.warning(__('extension %r has no setup() function; is it really '
                                   'a Sphinx extension module?'), extname)
-                metadata = {}  # type: Dict[str, Any]
+                metadata: Dict[str, Any] = {}
             else:
                 try:
                     metadata = setup(app)
@@ -491,7 +447,7 @@ class SphinxComponentRegistry:
                         __('The %s extension used by this project needs at least '
                            'Sphinx v%s; it therefore cannot be built with this '
                            'version.') % (extname, err)
-                    )
+                    ) from err
 
             if metadata is None:
                 metadata = {}
@@ -503,18 +459,25 @@ class SphinxComponentRegistry:
 
             app.extensions[extname] = Extension(extname, mod, **metadata)
 
-    def get_envversion(self, app):
-        # type: (Sphinx) -> Dict[str, str]
+    def get_envversion(self, app: "Sphinx") -> Dict[str, str]:
         from sphinx.environment import ENV_VERSION
         envversion = {ext.name: ext.metadata['env_version'] for ext in app.extensions.values()
                       if ext.metadata.get('env_version')}
         envversion['sphinx'] = ENV_VERSION
         return envversion
 
+    def get_publisher(self, app: "Sphinx", filetype: str) -> Publisher:
+        try:
+            return self.publishers[filetype]
+        except KeyError:
+            pass
+        publisher = create_publisher(app, filetype)
+        self.publishers[filetype] = publisher
+        return publisher
 
-def merge_source_suffix(app, config):
-    # type: (Sphinx, Config) -> None
-    """Merge source_suffix which specified by user and added by extensions."""
+
+def merge_source_suffix(app: "Sphinx", config: Config) -> None:
+    """Merge any user-specified source_suffix with any added by extensions."""
     for suffix, filetype in app.registry.source_suffix.items():
         if suffix not in app.config.source_suffix:
             app.config.source_suffix[suffix] = filetype
@@ -527,9 +490,8 @@ def merge_source_suffix(app, config):
     app.registry.source_suffix = app.config.source_suffix
 
 
-def setup(app):
-    # type: (Sphinx) -> Dict[str, Any]
-    app.connect('config-inited', merge_source_suffix)
+def setup(app: "Sphinx") -> Dict[str, Any]:
+    app.connect('config-inited', merge_source_suffix, priority=800)
 
     return {
         'version': 'builtin',
